/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet webserver with reset push button
 *
 * Chip type           : atmega88 or atmega168 or atmega328p or atmega644 with ENC28J60
 * Note: there is a version number in the text. Search for "ver."
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "websrv_help_functions.h"

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
// listen port for tcp/www:
static uint16_t mywwwport=80;
//
// bigger web-pages need a bigger buffer:
#define BUFFER_SIZE 450
static uint8_t buf[BUFFER_SIZE+1];
static volatile uint8_t stepcounter=0;
// global string buffer
#define STR_BUFFER_SIZE 24
static char strbuf[STR_BUFFER_SIZE+1];
//
// the password string (only the first 7 char checked), (only a-z,0-9,_ characters):
static char password[10]="secret"; // must not be longer than 9 char

// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

// takes a string of the form co?pw=xxx&rst=1 and analyse it
// return values: -1 not the reset page
//                -2 no rst command
//                -3 valid password
//                 1 rst=1 and passpword OK
int8_t analyse_get_url(char *str)
{
        if (strncmp("co",str,2)!=0){
                return(-1);
        }
        if (find_key_val(str,strbuf,STR_BUFFER_SIZE,"rst")){
                if (find_key_val(str,strbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(strbuf);
                        if (verify_password(strbuf)==0){
                                return(-3);
                        }else{
                                return(1);
                        }
                }else{
                        return(-3);
                }
        }
        // no rst
        return(-2);
}

// answer HTTP/1.0 301 Moved Permanently\r\nLocation: password/\r\n\r\n
// to redirect to the url ending in a slash
uint16_t moved_perm(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
        plen=fill_tcp_data(buf,plen,password);
        plen=fill_tcp_data_p(buf,plen,PSTR("/\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h1>301 Moved Permanently</h1>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("add a trailing slash to the url\n"));
        return(plen);
}


// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>reset/reboot button</h2>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/co method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\npw: <input type=password name=pw><input type=hidden name=rst value=1><input type=submit value=reboot></form><br>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr>ver. 2.16\n"));
        return(plen);
}

uint16_t print_webpage_confirm(uint8_t *buf)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2>\nReset executed. <a href=/>&gt;&gt;continue</a><br><hr>\n"));
        return(plen);
}

// timer interrupt, step counter and clear relay when at 2
ISR(TIMER1_COMPA_vect){
        if (stepcounter>1){
                stepcounter=0;
                PORTD &= ~(1<<PORTD7);// transistor off
        }
        if (stepcounter>0){
                stepcounter++;
        }
}


// Generate an interrup about ever 1/2s form the 12.5MHz system clock
// Since we have that 1024 prescaler we do not really generate a second
// (1.00000256000655361677s) 
void timer_init(void)
{
        /* write high byte first for 16 bit register access: */
        TCNT1H=0;  /* set counter to zero*/
        TCNT1L=0;
        // Mode 4 table 14-4 page 132. CTC mode and top in OCR1A
        // WGM13=0, WGM12=1, WGM11=0, WGM10=0
        TCCR1A=(0<<COM1B1)|(0<<COM1B0)|(0<<WGM11);
        TCCR1B=(1<<CS12)|(1<<CS10)|(1<<WGM12)|(0<<WGM13); // crystal clock/1024

        // divide crystal clock by 16+1:
        //ICR1H=0;
        //ICR1L=16; // top for counter
        // At what value to cause interrupt. You can use this for calibration
        // of the clock. Theoretical value for 25MHz: 6103=0x17 and 0xd7
        OCR1AH=0x17;
        OCR1AL=0xd7;
        // interrupt mask bit:
        TIMSK1 = (1 << OCIE1A);
}

int main(void){

        uint16_t plen;
        uint16_t dat_p;
        uint8_t i=0;
        int8_t cmd;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        /* enable PD2/INT0, as input */
        DDRD&= ~(1<<DDD2);

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        
        // LED
        /* enable PB1, LED as output */
        DDRB|= (1<<DDB1);

        /* set output to Vcc, LED off */
        PORTB|= (1<<PORTB1);

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PORTD7);// transistor off
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_loop_1(0); // 60us
        
        /* set output to GND, red LED on */
        PORTB &= ~(1<<PORTB1);
        i=1;

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);
        timer_init();

        sei(); // interrupt enable

        while(1){
                // handle ping and wait for a tcp packet:
                dat_p=packetloop_icmp_tcp(buf,enc28j60PacketReceive(BUFFER_SIZE, buf));
                /*plen will ne unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(dat_p==0){
                        continue;
                }
                        
                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                        // head, post and other methods:
                        //
                        // for possible status codes see:
                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"));
                        plen=print_webpage(buf);
                        goto SENDTCP;
                }
                // we chop off already the leading "/":
                cmd=analyse_get_url((char *)&(buf[dat_p+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (cmd==1){
                        PORTD|= (1<<PORTD7);// transistor on
                        stepcounter=1; // trigger counter to clear transistor 1 sec later
                        plen=print_webpage_confirm(buf);
                }else{
                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>&gt;&gt;continue</a>\n"));
                }

SENDTCP:
                www_server_reply(buf,plen); // send web page data
                // udp not implemented
        }
        return (0);
}
